#!/usr/local/bin/python3
#coding: utf-8
'''This is an executable wrapper for surgeo.'''

###############################################################################
# Main
###############################################################################

import argparse
import sys

import surgeo


def main(*args):
    '''This is the main application when running the program from a CLI.

    Args:
        --setup: (0 args) downloads and creates database for model creation
        --pipe: (0 args) takes stdin, processes, and sends to stdout
        --file: (2 args) takes 1. filepath input csv 2. filepath output csv
        --simple: (2 args) takes zip and surname, returns text string
        --complex: (2 args) takes zip and surname, returns detailed string
    Returns:
        --setup: None
        --pipe: long text string
        --file: None (output to csv file)
        --simple: text string ('White')
        --complex: long text string
    Raises:
        None

    '''

##### Parse arguments
    parsed_args = surgeo.utilities.get_parser_args()
##### Setup
    if parsed_args.setup:
        surgeo.data_setup(verbose=True)
##### Pipe
    if parsed_args.pipe:
        model = surgeo.SurgeoModel()
        try:
            while True:
                for line in sys.stdin:
                    try:
                        # Remove surrounding whitespace
                        line.strip()
                        zcta, surname = line.split()
                        result = model.race_data(zcta, surname)
                    except ValueError:
                        result = model.race_data('00000', 'BAD_NAME')
                    print(result.as_string)
        except EOFError:
            pass
##### Simple
    elif parsed_args.simple:
        model = surgeo.SurgeoModel()
        zcta = parsed_args.simple[0]
        surname = parsed_args.simple[1]
        race = model.guess_race(zcta, surname)
        print(race)
##### Complex
    elif parsed_args.complex:
        model = surgeo.SurgeoModel()
        zcta = parsed_args.complex[0]
        surname = parsed_args.complex[1]
        result = model.race_data(zcta, surname)
        print(result.as_string)
##### File
    elif parsed_args.file:
        model = surgeo.SurgeoModel()
        infile = parsed_args.file[0]
        outfile = parsed_args.file[1]
        model.process_csv(infile, outfile)
    elif not any([parsed_args.setup,
                  parsed_args.pipe,
                  parsed_args.simple,
                  parsed_args.complex,
                  parsed_args.file]):
        print('No arguments given. Try \'--help\'?')
        # GUI code to go here

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



'''surgeo is a Bayesian Improved Surname Geocoding model written in Python 3.
 ___ _   _ _ __ __ _  ___  ___
/ __| | | | '__/ _` |/ _ \/ _ \
\__ \ |_| | | | (_| |  __/ (_) |
|___/\__,_|_|  \__, |\___|\___/
               |___/

Python code by Theo Naunheim. Model created by Mark N. Elliot et al.

###############################################################################
License
###############################################################################

The MIT License (MIT)

Copyright (c) 2014 Theo Naunheim

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

'''

import sys


# Version check before imports. If not Python3, kill import
if sys.hexversion < 0x03000000:
    raise ImportError('Python version of 3.0 or higher required.')


# Put SurgeoModel, SurgeoResult, and SurgeoError in surgeo namespace
import surgeo
import surgeo.db
import surgeo.model
import surgeo.utilities

from surgeo.utilities.error_class import SurgeoError

from surgeo.utilities.surgeo_class import SurgeoModel
from surgeo.utilities.surgeo_class import SurgeoResult
from surgeo.utilities.surgeo_class import SurgeoErrorResult

from surgeo.utilities.sur_class import SurModel
from surgeo.utilities.sur_class import SurResult
from surgeo.utilities.sur_class import SurErrorResult

from surgeo.utilities.geo_class import GeoModel
from surgeo.utilities.geo_class import GeoResult
from surgeo.utilities.geo_class import GeoErrorResult


def data_setup(verbose=True):
    '''Downloads data needed to instantiate SurgeoModel.

    Args:
        verbose: True/False determines whether text is output.
    Returns:
        None
    Raises:
        None

    '''

    # Create necessary user data in home directory
    surgeo.utilities.setup_folder(verbose)
    # Setup surname db
    surgeo.db.setup_surname_table(verbose)
    # Setup race db
    surgeo.db.setup_geocode_table(verbose)
    # Reconstitute items supressed for confidentiality and index
    surgeo.db.reconstitute_data()
    # Remove zip and other unnecessary files
    surgeo.utilities.folder_cleanup()




import collections
import ftplib
import itertools
import os
import sqlite3
import sys
import time
import traceback
import zipfile


def setup_geocode_table(verbose):
    '''This sets up the geocoding database.

    Args:
        verbose: True/False for whether function outputs info.
    Returns:
        None
    Raises:
        None

    This function first downloads a geographic header file for each state. It
    then downloads file 00002 for each state. These are downloaded in zip
    format, and then unzipped. It then creates two tables, geocode_data and
    logical_race_data and populates the database. The geocode_data contains
    data for geographic areas. The logical record from the geocode_data
    directly correlates with a specific population in the geographic area,
    which is broken down by race.

    '''
    if verbose is True:
        sys.stdout.write('Downloading census files ... \t\t\n')
    # Created named tuple for organizing
    home_dir_path = os.path.expanduser("~")
    data_dir_path = os.path.join(home_dir_path, '.surgeo')
    # Download files from census server.
    ftp = ftplib.FTP('ftp.census.gov')
    ftp.login()
    ftp.cwd('census_2000/datasets/Summary_File_1')
    # List dir
    state_list = ftp.nlst()
    time.sleep(0)
    # Drop all elements prior to states
    state_list = itertools.dropwhile(lambda x: x != 'Alabama', state_list)
    # Make dropwhile object to list
    state_list = list(state_list)
    zip_files_downloaded = []
    for state in state_list:
        time.sleep(0)
        ftp.cwd('/')
        ftp.cwd(''.join(['census_2000/datasets/Summary_File_1',
                         '/',
                         state]))
        file_list = ftp.nlst()
        for item in file_list:
            time.sleep(0)
            if '00002_uf1.zip' in item or 'geo_uf1.zip' in item:
                if verbose is True:
                    print(''.join(['Downloading ', item]))
                file_path = os.path.join(data_dir_path, item)
                zip_files_downloaded.append(file_path)
                ftp.retrbinary('RETR ' + item, open(file_path, 'wb+').write)
    if verbose is True:
        sys.stdout.write('\t\t\t\t\tOK\n')
    # unzip files
    for zipfile_path in zip_files_downloaded:
        time.sleep(0)
        # Name of XXgeo_uf1.zip --> XXgeo.uf1
        # Name of XX00002_uf1.zip --> XX0000.uf1
        file_component = os.path.basename(zipfile_path).replace('.zip', '')
        file_component = file_component.replace('_', '.')
        if verbose is True:
            print('Writing {}'.format(file_component))
        dir_component = os.path.dirname(zipfile_path)
        # Zip file is now and iterator to save on ram.
        with zipfile.ZipFile(zipfile_path, 'r') as f:
            with f.open(file_component, 'r') as f2:
                with open(os.path.join(dir_component,
                          file_component),
                          'w+b') as f3:
                    for line in f2:
                        f3.write(line)
    # Now everything has been downloaded. Start commit to db
    try:
        db_path = os.path.join(os.path.expanduser('~'),
                               '.surgeo',
                               'census.db')
        connection = sqlite3.connect(db_path)
        cursor = connection.cursor()
        cursor.execute('''CREATE TABLE IF NOT EXISTS
                          geocode_data(id INTEGER PRIMARY KEY,
                          state TEXT, summary_level TEXT, logical_record TEXT,
                          zcta TEXT)''')
        cursor.execute('''CREATE TABLE IF NOT EXISTS logical_race_data(id
                          INTEGER PRIMARY KEY, state TEXT, logical_record TEXT,
                          num_white REAL, num_black REAL, num_ai REAL,
                          num_api REAL, num_hispanic REAL, num_multi REAL)''')
        # now start loading to db
        list_of_filenames = os.listdir(data_dir_path)
        number_of_filenames = len(list_of_filenames)
        for index, filename in enumerate(list_of_filenames):
            time.sleep(0)
            # First the geographic header file
            if 'geo.uf1' in filename:
                if verbose is True:
                    try:
                        last_write
                    except NameError:
                        last_write = 1
                    if index > last_write:
                        sys.stdout.write('\rWriting geoheader: {} of {}'
                                         .format(index,
                                                 number_of_filenames))
                        last_write = index
                file_path = os.path.join(data_dir_path,
                                         filename)
                #
                DESIRED_SUMMARY_LEVEL = '871'
                # Only latin1 appears to work, even thoug site specifies ascii
                with open(file_path, 'r', encoding='latin-1') as f3:
                    for line in f3:
                        time.sleep(0)
                        state = line[6:8]
                        summary_level = line[8:11]
                        logical_record = line[18:25]
                        zcta = line[160:165]
                        # Only ZCTA wide numbers considered
                        if not summary_level == DESIRED_SUMMARY_LEVEL:
                            continue
                        cursor.execute('''INSERT INTO geocode_data(id,
                                          state, summary_level, logical_record,
                                          zcta) VALUES(NULL, ?, ?, ?, ?)''',
                                       (state,
                                        summary_level,
                                        logical_record,
                                        zcta))
        if verbose is True:
            sys.stdout.write('\rWriting geoheader: {} of {}\n'.format
                            (number_of_filenames, number_of_filenames))
        for index, filename in enumerate(list_of_filenames):
            time.sleep(0)
            # First the geographic header file
            if '00002.uf1' in filename:
                if verbose is True:
                    try:
                        last_write
                        last_write = 1
                    except NameError:
                        last_write = 1
                    if index > last_write:
                        sys.stdout.write('\rWriting race data: {} of {}'
                                         .format(index,
                                                 number_of_filenames))
                        sys.stdout.flush()
                        last_write = index
                file_path = os.path.join(data_dir_path,
                                         filename)
                with open(file_path, 'r', encoding='latin-1') as f4:
                    for line in f4:
                        time.sleep(0)
                        state = line[5:7]
                        logical_record = line[15:22]
                        table_p8 = line.split(',')[86:103]
                        # Breaking up table p8
                        total_pop = table_p8[0]
                        total_not_hispanic = table_p8[1]
                        num_white = table_p8[2]
                        num_black = table_p8[3]
                        num_ai = table_p8[4]
                        num_asian = table_p8[5]
                        num_pacisland = table_p8[6]
                        num_api = str(int(num_asian) + int(num_pacisland))
                        num_other = table_p8[7]
                        num_multi = table_p8[8]
                        num_hispanic = table_p8[9]
                        cursor.execute('''INSERT INTO logical_race_data(
                                          id, state, logical_record,
                                          num_white, num_black,
                                          num_ai, num_api,
                                          num_hispanic, num_multi) VALUES(
                                          NULL, ?, ?, ?, ?, ?, ?, ?, ?)''',
                                      (state,
                                       logical_record,
                                       num_white,
                                       num_black,
                                       num_ai,
                                       num_api,
                                       num_multi,
                                       num_hispanic))
        if verbose is True:
            sys.stdout.write('\rWriting blockfile: {} of {}\n'
                             .format(number_of_filenames, number_of_filenames))
            sys.stdout.write('Creating indices ... \t\t\t')
            sys.stdout.flush()
        cursor.execute('''CREATE INDEX IF NOT EXISTS zcta_index ON
                          geocode_data(zcta)''')
        cursor.execute('''CREATE INDEX IF NOT EXISTS logical_record_index ON
                          logical_race_data(logical_record)''')
        sys.stdout.write('OK\n')
        # Now commit
        connection.commit()
        connection.close()
    except sqlite3.Error as e:
        traceback.print_exc()
        connection.rollback()
        connection.close()
        raise e




import decimal
import os
import sqlite3
import sys
import time
import traceback


def reconstitute_data():
    '''Go through each row. Fill in estimates for redacted items.

    Args:
        None
    Returns:
        None
    Raises:
        None

    It is a simple function that transforms the database. For confidentiality
    purposes, the Census Bureau scrubs certain data on race. This function
    reconstitutes that data. We take the total number of scrubbed items, and
    then divide it equally between the scrubbed categories. This yields an
    approximation.

    '''

    db_path = os.path.join(os.path.expanduser('~'),
                           '.surgeo',
                           'census.db')
    redacted_db = sqlite3.connect(db_path)
    try:
        cursor = redacted_db.cursor()
        altered_rows = []
        for row in cursor.execute('''SELECT * FROM surname_data'''):
            time.sleep(0)
            primary_key = row[0]
            surname = row[1]
            rank = row[2]
            count = row[3]
            prop1000k = row[4]
            cum_prop1000k = row[5]
            pctwhite = row[6]
            pctblack = row[7]
            pctapi = row[8]
            pctaian = row[9]
            pct2prace = row[10]
            # For some reason the last elemet often has a newline
            if type(row[11]) is str:
                pcthispanic = row[11].replace('\n', '')
            else:
                pcthispanic = row[11]
            # Per the study, the total number of redacted names is divided
            # by the number of redacted entries yielding an approximation.
            percentages = [pctwhite,
                           pctblack,
                           pctapi,
                           pctaian,
                           pct2prace,
                           pcthispanic]
            # Do not alter row unless it contains '(S)', which denotes redacted
            if not '(S)' in row:
                continue
            non_redacted = [x for x in percentages if not x == '(S)']
            redacted = [x for x in percentages if x == '(S)']
            non_redacted_percentage = sum(non_redacted)
            redacted_percentage = float(100) - (non_redacted_percentage)
            count_redacted_total = float(count) * redacted_percentage / 100
            count_per_redacted_item = round(count_redacted_total /
                                            len(redacted))
            # All redacted items set the same and added to list of altered rows
            for redacted_item in redacted:
                redacted_item = count_per_redacted_item
            # Add altered row to altered_rows
            altered_rows.append(row)
        for row in altered_rows:
            time.sleep(0)
            primary_key = row[0]
            cursor.execute('''DELETE FROM surname_data WHERE id=?''',
                           (primary_key,))
            cursor.execute('''INSERT INTO surname_data VALUES
                              (?,?,?,?,?,?,?,?,?,?,?,?)''', row)
        # Index via name to speed up searches
    except sqlite3.Error as e:
        traceback.print_exc()
        redacted_db.rollback()
        redacted_db.commit()
        raise e



'''This is the main class for using the model in other Python3 programs.

   Added to the surgeo namespace.'''

import csv
import io
import itertools
import operator
import os
import sqlite3
import sys

import surgeo

###############################################################################


class SurgeoModel(object):
    '''Contains data references and methods for running a BISG model.

    Attributes:
        self.db: an sqlite3 database connection shared for all methods
    Methods:
        guess_race: takes zip and surname and returns a string.
        race_data: takes zip and surname and returns a SurgeoResult object.
        process_csv: takes to paths. Reads path one. Processed result to path 2

    '''

    def __init__(self):
        # Load entire db to memory for performance DISK DATABSE
        db_path = os.path.join(os.path.expanduser('~'),
                               '.surgeo',
                               'census.db')
        if not os.path.exists(db_path):
            raise surgeo.SurgeoError('DB does not exist. Run surgeo.data'
                                     '_setup() or run program with '
                                     '\'--setup\' option.')
        self.db = sqlite3.connect(db_path)

    def guess_race(self, zcta, surname):
        '''zcta and surname go in and a simple race string comes out.

        Args:
            zcta: zip code, string or int
            surname: string
        Returns:
            result.probable_race: string
        Raises:
            None

        '''
        # Check for existence of zip code. Return bad result if bad.
        try:
            cursor = self.db.cursor()
            cursor.execute('''SELECT state, logical_record FROM
                           geocode_data WHERE zcta=?''', (zcta,))
            state, logical_record = cursor.fetchone()
            # Zip code is in database, so run it.
            result = surgeo.model.model1.run_model(zcta,
                                                   surname.upper(),
                                                   self.db)
        except TypeError:
            result = SurgeoErrorResult()
        return result.probable_race

    def race_data(self, zcta, surname):
        '''zcta and surname goes in, formatted SurgeoResult comes out.

        Args:
            zcta: zip code, string or int
            surname: string
        Returns:
            result: SurgeoResult object
        Raises:
            None

        '''

        # Check for existence of zip code. Return bad result if bad.
        try:
            cursor = self.db.cursor()
            cursor.execute('''SELECT state, logical_record FROM
                           geocode_data WHERE zcta=?''', (zcta,))
            state, logical_record = cursor.fetchone()
            # Zip code is in database, so run it.
            result = surgeo.model.model1.run_model(zcta,
                                                   surname.upper(),
                                                   self.db)
        except TypeError:
            result = SurgeoErrorResult()
        return result

    def process_csv(self,
                    filepath_in,
                    filepath_out,
                    verbose=True):
        '''This takes a csv filepath and creates new csv with race data.

        Args:
            filepath_in: file path of csv from which data is read
            filepath_out: file path of csv where data is written
            verbose: True/False which determines if updates are printed
        Returns:
            None
        Raises:
            None

        '''

        # Open file, determine if zip and name in header
        tempfile = open(filepath_in, 'rU')
        number_of_rows = len(tempfile.readlines())
        tempfile.close()
        with open(filepath_in, 'rU') as input_csv:
            csv_reader = csv.reader(input_csv)
            row_1 = next(csv_reader)
            number_of_columns = len(row_1)
            for index, item in enumerate(row_1):
                if type(item) is str:
                    if 'zip' in item.lower() or 'zcta' in item.lower():
                        zip_index = index
                    if 'last nam' in item.lower() or 'surname' in item.lower():
                        surname_index = index
            # zip/name index required. If no index for zip/name, raise error
            try:
                zip_index
                surname_index
            except NameError:
                raise SurgeoError('.csv row 1 lacks \'zip\' or ' +
                                  '\'surname\' fields.')
            # Go through all lines, amended lines with new data, and stage.
            # Writing to string buffer rather than file.
            line_buffer = io.StringIO()
            csv_writer = csv.writer(line_buffer)
            # Write header row first by splicing together
            chopped_header_row = row_1[:number_of_columns]
            header_remainder = ['surname', 'zip', 'probable_race',
                                'probable_race_percentage', 'hispanic',
                                'white', 'black', 'asian_or_pi',
                                'american_indian', 'multiracial']
            header_row = [item for item in itertools.chain(chopped_header_row,
                                                           header_remainder)]
            csv_writer.writerow(header_row)
            if verbose is True:
                sys.stdout.write('\rReading from {}\n'.format(filepath_in))
            # You already did a next() before on the iterator, line 2
            for index, entry in enumerate(csv_reader, start=1):
                surname = entry[surname_index]
                zcta = entry[zip_index]
                # Invoke object's race_data
                try:
                    cursor = self.db.cursor()
                    cursor.execute('''SELECT state, logical_record FROM
                                      geocode_data WHERE zcta=?''', (zcta,))
                    state, logical_record = cursor.fetchone()
                    # Zip code is in database, so run it.
                    result = surgeo.model.model1.run_model(zcta,
                                                           surname.upper(),
                                                           self.db)
                except TypeError:
                    result = SurgeoErrorResult()
                result_list = [result.surname,
                               result.zcta,
                               result.probable_race,
                               result.probable_race_percentage,
                               result.hispanic,
                               result.white,
                               result.black,
                               result.asian_or_pi,
                               result.american_indian,
                               result.multiracial]
                # Chop row in the event that rows are a different length
                chopped_row = entry[:number_of_columns]
                # Splice new row together
                new_row = [item for item in
                           itertools.chain(chopped_row, result_list)]
                csv_writer.writerow(new_row)
            if verbose is True:
                sys.stdout.write('Writing to {}.'.format(filepath_out))
            with open(filepath_out, 'w+') as f:
                f.write(line_buffer.getvalue())
            line_buffer.close()
            if verbose is True:
                sys.stdout.write('\nComplete.\n')


###############################################################################


class SurgeoResult(object):
    '''Result class containing BISG data.

    Attributes:
        self.surname: string
        self.zcta: string
        self.zip: string
        self.hispanic: float
        self.white: float
        self.black: float
        self.asian_or_pi: float
        self.american_indian: float
        self.multiracial: float
        self.probable_race: @property string
        self.probable_race_percentage: @property float
        self.as_string: @property string

    '''

    def __init__(self,
                 surname,
                 zcta,
                 hispanic,
                 white,
                 black,
                 asian_or_pi,
                 american_indian,
                 multiracial):
        self.surname = surname
        self.zcta = zcta
        self.zip = zcta
        self.hispanic = hispanic
        self.white = white
        self.black = black
        self.asian_or_pi = asian_or_pi
        self.american_indian = american_indian
        self.multiracial = multiracial

    @property
    def probable_race(self):
        rank_dict = {'Hispanic': self.hispanic,
                     'White': self.white,
                     'Black': self.black,
                     'Asian / Pacific Islander': self.asian_or_pi,
                     'American Indian / Alaskan Eskimo': self.american_indian,
                     'Multiracial': self.multiracial}
        most_probable_race = sorted(rank_dict.items(),
                                    key=operator.itemgetter(1),
                                    reverse=True)[0][0]
        return most_probable_race

    @property
    def probable_race_percentage(self):
        return max(self.hispanic,
                   self.white,
                   self.black,
                   self.asian_or_pi,
                   self.american_indian,
                   self.multiracial)

    @property
    def as_string(self):
        string = '\n'.join(['probable_race={}'.format(self.probable_race),
                            'probable_race_percent={}'.
                            format(self.probable_race_percentage),
                            'surname={}'.format(self.surname),
                            'zip={}'.format(self.zcta),
                            'hispanic={}'.format(self.hispanic),
                            'white={}'.format(self.white),
                            'black={}'.format(self.black),
                            'asian={}'.format(self.asian_or_pi),
                            'indian={}'.format(self.american_indian),
                            'multiracial={}'.format(self.multiracial)])
        return string

    @property
    def as_csv(self):
        original_string = str(self.as_string)
        original_as_list = original_string.split('\n')
        # Get rid of 'probable_race='
        original_as_list = [item.partition('=')[2] for item in
                            original_as_list]
        # Get rid of empties, and turn to csv
        original_as_list = [''.join(['\"', item, '\"']) for item in
                            original_as_list if item is not None]
        csv_string = ','.join(original_as_list) + '\n'
        return csv_string


class SurgeoErrorResult(object):
    '''Result class containing error data.

    Attributes:
        self.surname: string
        self.zcta: string
        self.zip: string
        self.hispanic: float
        self.white: float
        self.black: float
        self.asian_or_pi: float
        self.american_indian: float
        self.multiracial: float
        self.probable_race: @property string
        self.probable_race_percentage: @property float
        self.as_string: @property string

    '''

    def __init__(self):
        self.surname = 'Error'
        self.zcta = '00000'
        self.zip = '00000'
        self.hispanic = 0
        self.white = 0
        self.black = 0
        self.asian_or_pi = 0
        self.american_indian = 0
        self.multiracial = 0
        self.probable_race = 'Error'
        self.probable_race_percentage = 0
        self.as_string = '\n'.join(['probable_race=Error',
                                    'probable_race_percent=Error',
                                    'surname=Error',
                                    'zip=Error',
                                    'hispanic=Error',
                                    'white=Error',
                                    'black=Error',
                                    'asian=Error',
                                    'indian=Error',
                                    'multiracial=Error'])
